@@ -140,11 +140,62 @@
  *	Added a second patch to remove the pidfile as part of the
  *	termination cleanup sequence.  This minimizes the potential for
  *	conflicting pidfiles causing immediate termination at boot time.
+ *	
+ * Wed Aug 21 09:13:03 CDT 1996:  Dr. Wettstein
+ *	Added ability to reload static symbols and kernel module symbols
+ *      under control of SIGUSR1 and SIGUSR2 signals.
  *
- * Sun May 12 12:18:21 MET DST 1996:  Martin Schulze
- *	Corrected incorrect/insecure use of strpbrk for a not necessarily
- *	null-terminated buffer.  Used a patch from Chris Hanson
- *	(cph@martigny.ai.mit.edu), thanks.
+ *	Added -p switch to select 'paranoid' behavior with respect to the
+ *	loading of kernel module symbols.
+ *
+ *	Informative line now printed whenever a state change occurs due
+ *	to signal reception by the daemon.
+ *
+ *	Added the -i and -I command line switches to signal the currently
+ *	executing daemon.
+ *
+ * Tue Nov 19 10:15:36 PST 1996: Leland Olds <olds@eskimo.com>
+ *	Corrected vulnerability to buffer overruns by rewriting LogLine
+ *	routine.  Obscenely long kernel messages will now be broken up
+ *	into lines no longer than LOG_LINE_LENGTH.
+ *
+ *	The last version of LogLine was vulnerable to buffer overruns:
+ *	- Kernel messages longer than LOG_LINE_LENGTH caused a buffer
+ *	  overrun.
+ *	- If a line was determined to be shorter than LOG_LINE_LENGTH,
+ *	  the routine "ExpandKadds" could cause the line grow by
+ *	  an unknown amount and overrun a buffer.
+ *	I turned these routines into a little parsing state machine that
+ *	should not have these problems.
+ *
+ * Sun Jun 15 16:23:29 MET DST 1997: Michael Alan Dorman
+ *	Some more glibc patches made by <mdorman@debian.org>.
+ *
+ * Thu Aug 21 12:11:27 MET DST 1997: Martin Schulze <joey@infodrom.north.de>
+ *	Fixed little mistake which prevented klogd from accepting a
+ *	console log 
+ *
+ * Fri Jan  9 00:39:52 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Changed the behaviour of klogd when receiving a terminate
+ *	signal.  Now the program terminates immediately instead of
+ *	completing the receipt of a kernel message, i.e the read()
+ *	call.  The old behaveiour could result in klogd being
+ *	recognized as being undead, because it'll only die after a
+ *	message has been received.
+ *
+ * Fri Jan  9 11:03:48 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Corrected some code that caused klogd to dump core when
+ *	receiving messages containing '%', some of them exist in
+ *	2.1.78.  Thanks to Chu-yeon Park <kokids@doit.ajou.ac.kr> for
+ *	informing me.
+ *
+ * Fri Jan  9 23:38:19 CET 1998: Florian La Roche <florian@knorke.saar.de>
+ *	Added -x switch to omit EIP translation and System.map evaluation.
+ *
+ * Sun Jan 25 20:47:46 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	As the bug covering the %'s introduced a problem with
+ *	unevaluated priorities I've worked out a real fix that strips
+ *	%'s to an even number which is harmless for printf.
  */
 
 
@@ -154,17 +205,26 @@
 #include <errno.h>
 #include <sys/fcntl.h>
 #include <sys/stat.h>
+#if !defined(__GLIBC__)
 #include <linux/time.h>
+#endif /* __GLIBC__ */
 #include <stdarg.h>
 #include <paths.h>
+#include <stdlib.h>
 #include "klogd.h"
+#include "ksyms.h"
 #include "pidfile.h"
 #include "version.h"
 
 #define __LIBRARY__
 #include <linux/unistd.h>
-#define __NR_sys_syslog __NR_syslog
-_syscall3(int,sys_syslog,int, type, char *, buf, int, len);
+#if !defined(__GLIBC__)
+# define __NR_ksyslog __NR_syslog
+_syscall3(int,ksyslog,int, type, char *, buf, int, len);
+#else
+#include <sys/klog.h>
+#define ksyslog klogctl
+#endif
 
 #define LOG_BUFFER_SIZE 4096
 #define LOG_LINE_LENGTH 1024
@@ -179,13 +239,16 @@ static int	kmsg,
 		change_state = 0,
 		terminate = 0,
 		caught_TSTP = 0,
+		reload_symbols = 0,
 		console_log_level = 6;
 
 static int	use_syscall = 0,
 		one_shot = 0,
-		NoFork = 0;	/* don't fork - don't run in daemon mode */
+		symbol_lookup = 1,
+		no_fork = 0;	/* don't fork - don't run in daemon mode */
 
-static char log_buffer[LOG_BUFFER_SIZE];
+static char	*symfile = (char *) 0,
+		log_buffer[LOG_BUFFER_SIZE];
 
 static FILE *output_file = (FILE *) 0;
 
@@ -195,12 +258,15 @@ int debugging = 0;
 
 
 /* Function prototypes. */
-extern int sys_syslog(int type, char *buf, int len);
+extern int ksyslog(int type, char *buf, int len);
 static void CloseLogSrc(void);
 extern void restart(int sig);
 extern void stop_logging(int sig);
 extern void stop_daemon(int sig);
+extern void reload_daemon(int sig);
 static void Terminate(void);
+static void SignalDaemon(int);
+static void ReloadSymbols(void);
 static void ChangeLogging(void);
 static enum LOGSRC GetKernelLogSrc(void);
 static void LogLine(char *ptr, int len);
@@ -213,14 +279,14 @@ static void CloseLogSrc()
 
 {
 	/* Turn on logging of messages to console. */
-  	sys_syslog(7, NULL, 0);
+  	ksyslog(7, NULL, 0);
   
         /* Shutdown the log sources. */
 	switch ( logsrc )
 	{
 	    case kernel:
-		sys_syslog(0, 0, 0);
-		Syslog(LOG_INFO, "Kernel logging (sys_syslog) stopped.");
+		ksyslog(0, 0, 0);
+		Syslog(LOG_INFO, "Kernel logging (ksyslog) stopped.");
 		break;
             case proc:
 		close(kmsg);
@@ -264,9 +330,29 @@ void stop_daemon(sig)
 
 	int sig;
 
+{
+	Terminate();
+	return;
+}
+
+
+void reload_daemon(sig)
+
+     int sig;
+
 {
 	change_state = 1;
-	terminate = 1;
+	reload_symbols = 1;
+
+
+	if ( sig == SIGUSR2 )
+	{
+		++reload_symbols;
+		signal(SIGUSR2, reload_daemon);
+	}
+	else
+		signal(SIGUSR1, reload_daemon);
+		
 	return;
 }
 
@@ -284,7 +370,31 @@ static void Terminate()
 	exit(1);
 }
 
-	
+static void SignalDaemon(sig)
+
+     int sig;
+
+{
+	auto int pid = check_pid(PidFile);
+
+	kill(pid, sig);
+	return;
+}
+
+
+static void ReloadSymbols()
+
+{
+	if (symbol_lookup) {
+		if ( reload_symbols > 1 )
+			InitKsyms(symfile);
+		InitMsyms();
+	}
+	reload_symbols = change_state = 0;
+	return;
+}
+
+
 static void ChangeLogging(void)
 
 {
@@ -292,6 +402,17 @@ static void ChangeLogging(void)
 	if ( terminate == 1 )
 		Terminate();
 
+	/* Indicate that something is happening. */
+	Syslog(LOG_INFO, "klogd %s-%s, ---------- state change ----------\n", \
+	       VERSION, PATCHLEVEL);
+
+	/* Reload symbols. */
+	if ( reload_symbols > 0 )
+	{
+		ReloadSymbols();
+		return;
+	}
+
 	/* Stop kernel logging. */
 	if ( caught_TSTP == 1 )
 	{
@@ -330,7 +451,7 @@ static enum LOGSRC GetKernelLogSrc(void)
 
 
 	/* Set level of kernel console messaging.. */
-	if ( (sys_syslog(8, NULL, console_log_level) < 0) && \
+	if ( (ksyslog(8, NULL, console_log_level) < 0) && \
 	     (errno == EINVAL) )
 	{
 		/*
@@ -341,7 +462,7 @@ static enum LOGSRC GetKernelLogSrc(void)
 		 */
 		Syslog(LOG_WARNING, "Cannot set console log level - disabling "
 			      "console output.");
-		sys_syslog(6, NULL, 0);
+		ksyslog(6, NULL, 0);
 	}
 	
 
@@ -353,16 +474,17 @@ static enum LOGSRC GetKernelLogSrc(void)
 	    ((stat(_PATH_KLOG, &sb) < 0) && (errno == ENOENT)) )
 	{
 	  	/* Initialize kernel logging. */
-	  	sys_syslog(1, NULL, 0);
-		Syslog(LOG_INFO, "klogd %s-%s, log source = sys_syslog "
+	  	ksyslog(1, NULL, 0);
+		Syslog(LOG_INFO, "klogd %s-%s, log source = ksyslog "
 		       "started.", VERSION, PATCHLEVEL);
 		return(kernel);
 	}
 	
 	if ( (kmsg = open(_PATH_KLOG, O_RDONLY)) < 0 )
 	{
-		fputs("klogd: Cannot open proc file system.", stderr);
-		sys_syslog(7, NULL, 0);
+		fprintf(stderr, "klogd: Cannot open proc file system, " \
+			"%d - %s.\n", errno, strerror(errno));
+		ksyslog(7, NULL, 0);
 		exit(1);
 	}
 
@@ -381,7 +503,7 @@ extern void Syslog(int priority, char *fmt, ...)
 	{
 		fputs("Logging line:\n", stderr);
 		fprintf(stderr, "\tLine: %s\n", fmt);
-		fprintf(stderr, "\tPriority: %c\n", *(fmt+1));
+		fprintf(stderr, "\tPriority: %d\n", priority);
 	}
 
 	/* Handle output to a file. */
@@ -392,7 +514,8 @@ extern void Syslog(int priority, char *fmt, ...)
 		va_end(ap);
 		fputc('\n', output_file);
 		fflush(output_file);
-		fsync(fileno(output_file));
+		if (!one_shot)
+			fsync(fileno(output_file));
 		return;
 	}
 	
@@ -436,79 +559,244 @@ extern void Syslog(int priority, char *fmt, ...)
 	return;
 }
 
-	
-static void LogLine(char *ptr, int len)
 
+/*
+ *     Copy characters from ptr to line until a char in the delim
+ *     string is encountered or until min( space, len ) chars have
+ *     been copied.
+ *
+ *     Returns the actual number of chars copied.
+ */
+static int copyin( char *line,      int space,
+                   const char *ptr, int len,
+                   const char *delim )
 {
-	auto int idx = 0;
-	static int index = 0;
-	auto char *nl;
-	auto char *pend = ptr + len;
-	static char line[LOG_LINE_LENGTH],
-		    eline[LOG_LINE_LENGTH];
+    auto int i;
+    auto int count;
 
+    count = len < space ? len : space;
 
-	if ( debugging && (len != 0) )
-	{
-		fprintf(stderr, "Log buffer contains: %d characters.\n", len);
-		fprintf(stderr, "Line buffer contains: %d characters.\n", \
-			index);
-		while ( idx <= len )
-		{
-			fprintf(stderr, "Character #%d - %d:%c\n", idx, \
-				ptr[idx], ptr[idx]);
-			++idx;
-		}
-		if ( index != 0 )
-		{
-			fputs("Line buffer contains an unterminated line:\n", \
-			      stderr);
-			fprintf(stderr, "\tCount: %d\n", index);
-			fprintf(stderr, "%s\n\n", line);
-		}
-	}
+    for(i=0; i<count && !strchr(delim, *ptr); i++ ) { *line++ = *ptr++; }
 
-	if ( index == 0 )
-		memset(line, '\0', sizeof(line));
-	
-	while (len) {
-		for (nl = ptr; nl < pend; nl += 1)
-			if ((*nl == '\n') || (*nl == '\r'))
-				break;
-		if (nl != pend) {
-			len -= nl - ptr + 1;
-			strncat(line, ptr, nl - ptr);
-			ptr = nl + 1;
-			/* Check for empty log line (may be produced if 
-			   kernel messages have multiple terminators, eg.
-			   \n\r) */
-			if ( (*line != '\n') && (*line != '\r') )
-			{
-				memset(eline, '\0', sizeof(eline));
-				ExpandKadds(line, eline);
-				Syslog(LOG_INFO, eline);
-			}
-			index = 0;
-			memset(line, '\0', sizeof(line));
-		 }
-		 else
-		 {
-			 if ( debugging )
-			 {
-				 fputs("No terminator - leftover:\n", stderr);
-				 fprintf(stderr, "\tCharacters: %d\n", len);
-				 fprintf(stderr, "\tIndex: %d\n", index);
-				 fputs("\tLine: ", stderr);
-				 fprintf(stderr, "%s\n", line);
-			 }
-			 
-			strncat(line, ptr, len);
-			index += len;
-			len = 0;
-		}
-	}
+    return( i );
+}
 
-	return;
+/*
+ * Messages are separated by "\n".  Messages longer than
+ * LOG_LINE_LENGTH are broken up.
+ *
+ * Kernel symbols show up in the input buffer as : "[<aaaaaa>]",
+ * where "aaaaaa" is the address.  These are replaced with
+ * "[symbolname+offset/size]" in the output line - symbolname,
+ * offset, and size come from the kernel symbol table.
+ *
+ * If a kernel symbol happens to fall at the end of a message close
+ * in length to LOG_LINE_LENGTH, the symbol will not be expanded.
+ * (This should never happen, since the kernel should never generate
+ * messages that long.
+ */
+static void LogLine(char *ptr, int len)
+{
+    enum parse_state_enum {
+        PARSING_TEXT,
+        PARSING_SYMSTART,      /* at < */
+        PARSING_SYMBOL,        
+        PARSING_SYMEND         /* at ] */
+    };
+
+    static char line_buff[LOG_LINE_LENGTH];
+
+    static char *line                        =line_buff;
+    static enum parse_state_enum parse_state = PARSING_TEXT;
+    static int space                         = sizeof(line_buff)-1;
+
+    static char *sym_start;            /* points at the '<' of a symbol */
+
+    auto   int delta = 0;              /* number of chars copied        */
+
+    while( len > 0 )
+    {
+        if( space == 0 )    /* line buffer is full */
+        {
+            /*
+            ** Line too long.  Start a new line.
+            */
+            *line = 0;   /* force null terminator */
+
+	    if ( debugging )
+	    {
+		fputs("Line buffer full:\n", stderr);
+		fprintf(stderr, "\tLine: %s\n", line);
+	    }
+
+            Syslog( LOG_INFO, line_buff );
+            line  = line_buff;
+            space = sizeof(line_buff)-1;
+            parse_state = PARSING_TEXT;
+        }
+
+        switch( parse_state )
+        {
+        case PARSING_TEXT:
+               delta = copyin( line, space, ptr, len, "\n[%" );
+               line  += delta;
+               ptr   += delta;
+               space -= delta;
+               len   -= delta;
+
+               if( space == 0 || len == 0 )
+               {
+		  break;  /* full line_buff or end of input buffer */
+               }
+
+               if( *ptr == '\n' )  /* newline */
+               {
+                  *line++ = *ptr++;  /* copy it in */
+                  space -= 1;
+                  len   -= 1;
+
+                  *line = 0;  /* force null terminator */
+	          Syslog( LOG_INFO, line_buff );
+                  line  = line_buff;
+                  space = sizeof(line_buff)-1;
+                  break;
+               }
+               if( *ptr == '[' )   /* possible kernel symbol */
+               {
+                  *line++ = *ptr++;
+                  space -= 1;
+                  len   -= 1;
+                  parse_state = PARSING_SYMSTART;      /* at < */
+                  break;
+               }
+               if( *ptr == '%' )   /* dangerous printf marker */
+	       {
+		   delta = 0;
+		   while (len && *ptr == '%')
+		   {
+		       *line++ = *ptr++;	/* copy it in */
+		       space -= 1;
+		       len   -= 1;
+		       delta++;
+		   }
+		   if (delta % 2)	/* odd amount of %'s */
+		   {
+		       if (space)
+		       {
+			   *line++ = '%';	/* so simply add one */
+			   space -= 1;
+		       }
+		       else 
+		       {
+			   *line++ = '\0';	/* remove the last one / terminate the string */
+		       }
+
+		   }
+	       }
+               break;
+        
+        case PARSING_SYMSTART:
+               if( *ptr != '<' )
+               {
+                  parse_state = PARSING_TEXT;        /* not a symbol */
+                  break;
+               }
+
+               /*
+               ** Save this character for now.  If this turns out to
+               ** be a valid symbol, this char will be replaced later.
+               ** If not, we'll just leave it there.
+               */
+
+               sym_start = line; /* this will point at the '<' */
+
+               *line++ = *ptr++;
+               space -= 1;
+               len   -= 1;
+               parse_state = PARSING_SYMBOL;     /* symbol... */
+               break;
+
+        case PARSING_SYMBOL:
+               delta = copyin( line, space, ptr, len, ">\n[" );
+               line  += delta;
+               ptr   += delta;
+               space -= delta;
+               len   -= delta;
+               if( space == 0 || len == 0 )
+               {
+                  break;  /* full line_buff or end of input buffer */
+               }
+               if( *ptr != '>' )
+               {
+                  parse_state = PARSING_TEXT;
+                  break;
+               }
+
+               *line++ = *ptr++;  /* copy the '>' */
+               space -= 1;
+               len   -= 1;
+
+               parse_state = PARSING_SYMEND;
+
+               break;
+
+        case PARSING_SYMEND:
+               if( *ptr != ']' )
+               {
+                  parse_state = PARSING_TEXT;        /* not a symbol */
+                  break;
+               }
+
+               /*
+               ** It's really a symbol!  Replace address with the
+               ** symbol text.
+               */
+           {
+	       auto int sym_space;
+
+	       auto int value;
+	       auto struct symbol sym;
+	       auto char *symbol;
+
+               *(line-1) = 0;    /* null terminate the address string */
+               value  = strtol(sym_start+1, (char **) 0, 16);
+               *(line-1) = '>';  /* put back delim */
+
+               symbol = LookupSymbol(value, &sym);
+               if ( !symbol_lookup || symbol == (char *) 0 )
+               {
+                  parse_state = PARSING_TEXT;
+                  break;
+               }
+
+               /*
+               ** verify there is room in the line buffer
+               */
+               sym_space = space + ( line - sym_start );
+               if( sym_space < strlen(symbol) + 30 ) /*(30 should be overkill)*/
+               {
+                  parse_state = PARSING_TEXT;  /* not enough space */
+                  break;
+               }
+
+               delta = sprintf( sym_start, "%s+%d/%d]",
+                                symbol, sym.offset, sym.size );
+
+               space = sym_space + delta;
+               line  = sym_start + delta;
+           }
+               ptr++;
+               len--;
+               parse_state = PARSING_TEXT;
+               break;
+
+        default: /* Can't get here! */
+               parse_state = PARSING_TEXT;
+
+        }
+    }
+
+    return;
 }
 
 
@@ -524,12 +812,12 @@ static void LogKernelLine(void)
 	 * messages into this fresh buffer.
 	 */
 	memset(log_buffer, '\0', sizeof(log_buffer));
-	if ( (rdcnt = sys_syslog(2, log_buffer, sizeof(log_buffer))) < 0 )
+	if ( (rdcnt = ksyslog(2, log_buffer, sizeof(log_buffer))) < 0 )
 	{
 		if ( errno == EINTR )
 			return;
-		fprintf(stderr, "Error return from sys_sycall: %d - %s\n", \
-			errno, strerror(errno));
+		fprintf(stderr, "klogd: Error return from sys_sycall: " \
+			"%d - %s\n", errno, strerror(errno));
 	}
 	
 	LogLine(log_buffer, rdcnt);
@@ -553,7 +841,8 @@ static void LogProcLine(void)
 	{
 		if ( errno == EINTR )
 			return;
-		Syslog(LOG_ERR, "Cannot read proc file system.");
+		Syslog(LOG_ERR, "Cannot read proc file system: %d - %s.", \
+		       errno, strerror(errno));
 	}
 	
 	LogLine(log_buffer, rdcnt);
@@ -569,14 +858,15 @@ int main(argc, argv)
 	char *argv[];
 
 {
-	auto int ch, use_output = 0;
+	auto int	ch,
+			use_output = 0;
 
-	auto char	*symfile = (char *) 0,
-			*log_level = (char *) 0,
+	auto char	*log_level = (char *) 0,
 			*output = (char *) 0;
 
+	chdir ("/");
 	/* Parse the command-line. */
-	while ((ch = getopt(argc, argv, "c:df:k:nosv")) != EOF)
+	while ((ch = getopt(argc, argv, "c:df:iIk:nopsvx")) != EOF)
 		switch((char)ch)
 		{
 		    case 'c':		/* Set console message level. */
@@ -589,21 +879,33 @@ int main(argc, argv)
 			output = optarg;
 			use_output++;
 			break;
+		    case 'i':		/* Reload module symbols. */
+			SignalDaemon(SIGUSR1);
+			return(0);
+		    case 'I':
+			SignalDaemon(SIGUSR2);
+			return(0);
 		    case 'k':		/* Kernel symbol file. */
 			symfile = optarg;
 			break;
 		    case 'n':		/* don't fork */
-			NoFork++;
+			no_fork++;
 			break;
 		    case 'o':		/* One-shot mode. */
 			one_shot = 1;
 			break;
+		    case 'p':
+			SetParanoiaLevel(1);	/* Load symbols on oops. */
+			break;	
 		    case 's':		/* Use syscall interface. */
 			use_syscall = 1;
 			break;
 		    case 'v':
 			printf("klogd %s-%s\n", VERSION, PATCHLEVEL);
 			exit (1);
+		    case 'x':
+			symbol_lookup = 0;
+			break;
 		}
 
 
@@ -611,7 +913,7 @@ int main(argc, argv)
 	if ( log_level != (char *) 0 )
 	{
 		if ( (strlen(log_level) > 1) || \
-		     (strchr("1234567", *log_level) == (char *) 0) )
+		     (strchr("12345678", *log_level) == (char *) 0) )
 		{
 			fprintf(stderr, "klogd: Invalid console logging "
 				"level <%s> specified.\n", log_level);
@@ -632,7 +934,7 @@ int main(argc, argv)
 	 * not disabled with the command line argument and there's no
 	 * such process running.
 	 */
-	if ( (!one_shot) && (!NoFork) )
+	if ( (!one_shot) && (!no_fork) )
 	{
 		if (!check_pid(PidFile))
 		{
@@ -685,6 +987,8 @@ int main(argc, argv)
 	signal(SIGHUP, stop_daemon);
 	signal(SIGTSTP, stop_logging);
 	signal(SIGCONT, restart);
+	signal(SIGUSR1, reload_daemon);
+	signal(SIGUSR2, reload_daemon);
 
 
 	/* Open outputs. */
@@ -694,8 +998,8 @@ int main(argc, argv)
 			output_file = stdout;
 		else if ( (output_file = fopen(output, "w")) == (FILE *) 0 )
 		{
-			fprintf(stderr, "klogd: Cannot open output file %s - "\
-				"%s\n", output, strerror(errno));
+			fprintf(stderr, "klogd: Cannot open output file " \
+				"%s - %s\n", output, strerror(errno));
 			return(1);
 		}
 	}
@@ -706,7 +1010,10 @@ int main(argc, argv)
 	/* Handle one-shot logging. */
 	if ( one_shot )
 	{
-		InitKsyms(symfile);
+		if (symbol_lookup) {
+			InitKsyms(symfile);
+			InitMsyms();
+		}
 		if ( (logsrc = GetKernelLogSrc()) == kernel )
 			LogKernelLine();
 		else
@@ -719,7 +1026,10 @@ int main(argc, argv)
 	sleep(KLOGD_DELAY);
 #endif
 	logsrc = GetKernelLogSrc();
-	InitKsyms(symfile);
+	if (symbol_lookup) {
+		InitKsyms(symfile);
+		InitMsyms();
+	}
 
         /* The main loop. */
 	while (1)
@@ -740,3 +1050,10 @@ int main(argc, argv)
 		}
 	}
 }
+/*
+ * Local variables:
+ *  c-indent-level: 4
+ *  c-basic-offset: 4
+ *  tab-width: 8
+ * End:
+ */
