@@ -208,19 +208,19 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
  *	reception on with the "-r" option.
  *
  *	Not defining SYSLOG_INET will result in not doing any network
- *	activity, i.e. not sending or receiving messages. I changed
+ *	activity, i.e. not sending or receiving messages.  I changed
  *	this because the old idea is implemented with the "-r" option
  *	and the old thing didn't work anyway.
  *
  * Thu Oct 26 13:14:06 MET 1995:  Martin Schulze
- *	Added another logfile type F_FORW_UNKN. The problem I ran into
+ *	Added another logfile type F_FORW_UNKN.  The problem I ran into
  *	was a name server that runs on my machine and a forwarder of
- *	kern.crit to another host. The hosts address can only be
- *	fetched using the nameserver. But named is started after
+ *	kern.crit to another host.  The hosts address can only be
+ *	fetched using the nameserver.  But named is started after
  *	syslogd, so syslogd complained.
  *
  *	This logfile type will retry to get the address of the
- *	hostname ten times and then complain. This should be enough to
+ *	hostname ten times and then complain.  This should be enough to
  *	get the named up and running during boot sequence.
  *
  * Fri Oct 27 14:08:15 1995:  Dr. Wettstein
@@ -270,17 +270,87 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
  *
  *	Minor code cleanups.
  *
+ * Thu May  2 15:15:33 CDT 1996:  Dr. Wettstein
+ *	Fixed bug in init function which resulted in file descripters
+ *	being orphaned when syslogd process was re-initialized with SIGHUP
+ *	signal.  Thanks to Edvard Tuinder
+ *	(Edvard.Tuinder@praseodymium.cistron.nl) for putting me on the
+ *	trail of this bug.  I am amazed that we didn't catch this one
+ *	before now.
+ *
  * Tue May 14 00:03:35 MET DST 1996:  Martin Schulze
  *	Corrected a mistake that causes the syslogd to stop logging at
- *	some virtual consoles under Linux. This was causes by checking
+ *	some virtual consoles under Linux. This was caused by checking
  *	the wrong error code. Thanks to Michael Nonweiler
  *	<mrn20@hermes.cam.ac.uk> for sending me a patch.
  *
+ * Mon May 20 13:29:32 MET DST 1996:  Miquel van Smoorenburg <miquels@cistron.nl>
+ *	Added continuation line supported and fixed a bug in
+ *	the init() code.
+ *
  * Tue May 28 00:58:45 MET DST 1996:  Martin Schulze
  *	Corrected behaviour of blocking pipes - i.e. the whole system
  *	hung.  Michael Nonweiler <mrn20@hermes.cam.ac.uk> has sent us
  *	a patch to correct this.  A new logfile type F_PIPE has been
  *	introduced.
+ *
+ * Mon Feb 3 10:12:15 MET DST 1997:  Martin Schulze
+ *	Corrected behaviour of logfiles if the file can't be opened.
+ *	There was a bug that causes syslogd to try to log into non
+ *	existing files which ate cpu power.
+ *
+ * Sun Feb 9 03:22:12 MET DST 1997:  Martin Schulze
+ *	Modified syslogd.c to not kill itself which confuses bash 2.0.
+ *
+ * Mon Feb 10 00:09:11 MET DST 1997:  Martin Schulze
+ *	Improved debug code to decode the numeric facility/priority
+ *	pair into textual information.
+ *
+ * Tue Jun 10 12:35:10 MET DST 1997:  Martin Schulze
+ *	Corrected freeing of logfiles.  Thanks to Jos Vos <jos@xos.nl>
+ *	for reporting the bug and sending an idea to fix the problem.
+ *
+ * Tue Jun 10 12:51:41 MET DST 1997:  Martin Schulze
+ *	Removed sleep(10) from parent process.  This has caused a slow
+ *	startup in former times - and I don't see any reason for this.
+ *
+ * Sun Jun 15 16:23:29 MET DST 1997: Michael Alan Dorman
+ *	Some more glibc patches made by <mdorman@debian.org>.
+ *
+ * Thu Jan  1 16:04:52 CET 1998: Martin Schulze <joey@infodrom.north.de
+ *	Applied patch from Herbert Thielen <Herbert.Thielen@lpr.e-technik.tu-muenchen.de>.
+ *	This included some balance parentheses for emacs and a bug in
+ *	the exclamation mark handling.
+ *
+ *	Fixed small bug which caused syslogd to write messages to the
+ *	wrong logfile under some very rare conditions.  Thanks to
+ *	Herbert Xu <herbert@gondor.apana.org.au> for fiddling this out.
+ *
+ * Thu Jan  8 22:46:35 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Reworked one line of the above patch as it prevented syslogd
+ *	from binding the socket with the result that no messages were
+ *	forwarded to other hosts.
+ *
+ * Sat Jan 10 01:33:06 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Fixed small bugs in F_FORW_UNKN meachanism.  Thanks to Torsten
+ *	Neumann <torsten@londo.rhein-main.de> for pointing me to it.
+ *
+ * Mon Jan 12 19:50:58 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Modified debug output concerning remote receiption.
+ *
+ * Mon Feb 23 23:32:35 CET 1998: Topi Miettinen <Topi.Miettinen@ml.tele.fi>
+ *	Re-worked handling of Unix and UDP sockets to support closing /
+ *	opening of them in order to have it open only if it is needed
+ *	either for forwarding to a remote host or by receiption from
+ *	the network.
+ *
+ * Wed Feb 25 10:54:09 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Fixed little comparison mistake that prevented the MARK
+ *	feature to work properly.
+ *
+ * Wed Feb 25 13:21:44 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Corrected Topi's patch as it prevented forwarding during
+ *	startup due to an unknown LogPort.
  */
 
 
@@ -290,6 +360,8 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
 #define DEFSPRI		(LOG_KERN|LOG_CRIT)
 #define TIMERINTVL	30		/* interval for checking flush, mark */
 
+#define CONT_LINE	1		/* Allow continuation lines */
+
 #include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -302,6 +374,7 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
 #include <setjmp.h>
 #include <stdarg.h>
 
+#define SYSLOG_NAMES
 #include <sys/syslog.h>
 #include <sys/param.h>
 #include <sys/errno.h>
@@ -390,7 +463,7 @@ char	ctty[] = _PATH_CONSOLE;
 
 char	**parts;
 
-int inetm = 0, funix = 0;
+int inetm = 0, funix = -1;
 static int debugging_on = 0;
 static int nlogs = -1;
 static int restart = 0;
@@ -546,7 +619,7 @@ int	Debug;			/* debug flag */
 char	LocalHostName[MAXHOSTNAMELEN+1];	/* our hostname */
 char	*LocalDomain;		/* our local domain name */
 int	InetInuse = 0;		/* non-zero if INET sockets are being used */
-int	finet;			/* Internet datagram socket */
+int	finet = -1;		/* Internet datagram socket */
 int	LogPort;		/* port number for INET connections */
 int	Initialized = 0;	/* set when we have initialized ourselves */
 int	MarkInterval = 20 * 60;	/* interval between marks in seconds */
@@ -559,7 +632,9 @@ int	NoHops = 1;		/* Can we bounce syslog messages through an
 				   intermediate host. */
 
 extern	int errno, sys_nerr;
+#if !defined(__GLIBC__)
 extern	char *sys_errlist[];
+#endif /* __GLIBC__ */
 extern	char *ctime(), *index();
 
 /* Function prototypes. */
@@ -583,10 +658,19 @@ void die(int sig);
 void init();
 void cfline(char *line, register struct filed *f);
 int decode(char *name, struct code *codetab);
+#if defined(__GLIBC__)
+#define dprintf mydprintf
+#endif /* __GLIBC__ */
 static void dprintf(char *, ...);
 static void allocate_log(void);
 void sighup_handler();
 
+#ifdef SYSLOG_UNIXAF
+static int create_unix_socket();
+#endif
+#ifdef SYSLOG_INET
+static int create_inet_socket();
+#endif
 
 int main(argc, argv)
 	int argc;
@@ -594,15 +678,20 @@ int main(argc, argv)
 {
 	register int i;
 	register char *p;
+#if !defined(__GLIBC__)
 	int len, num_fds;
+#else /* __GLIBC__ */
+	size_t len;
+	int num_fds;
+#endif /* __GLIBC__ */
 	fd_set unixm, readfds;
 
 	int	fd;
 #ifdef SYSLOG_UNIXAF
-	struct sockaddr_un sunx, fromunix;
+	struct sockaddr_un fromunix;
 #endif
 #ifdef  SYSLOG_INET
-	struct sockaddr_in sin, frominet;
+	struct sockaddr_in frominet;
 	char *from;
 #endif
 	int ch;
@@ -612,8 +701,7 @@ int main(argc, argv)
 	extern int optind;
 	extern char *optarg;
 
-	int quitpid = 0;
-
+	chdir ("/");
 	while ((ch = getopt(argc, argv, "dhf:l:m:np:rs:v")) != EOF)
 		switch((char)ch) {
 		case 'd':		/* debug */
@@ -655,7 +743,7 @@ int main(argc, argv)
 			break;
 		case 'v':
 			printf("syslogd %s-%s\n", VERSION, PATCHLEVEL);
-			exit (1);
+			exit (0);
 		case '?':
 		default:
 			usage();
@@ -668,13 +756,11 @@ int main(argc, argv)
 		dprintf("Checking pidfile.\n");
 		if (!check_pid(PidFile))
 		{
-			quitpid = getpid();
 			if (fork())
-			{
-				/* We try to wait the end of initialization */
-				sleep(10);
+				/*
+				 * Parent process
+				 */
 				exit(0);
-			}
 			num_fds = getdtablesize();
 			for (i= 0; i < num_fds; i++)
 				(void) close(i);
@@ -728,16 +814,20 @@ int main(argc, argv)
 		 * should return the simple hostname or the fqdn. A
 		 * good piece of software should be aware of both and
 		 * we want to distribute good software.  Joey
+		 *
+		 * Good software also always checks its return values...
+		 * If syslogd starts up before DNS is up & /etc/hosts
+		 * doesn't have LocalHostName listed, gethostbyname will
+		 * return NULL. 
 		 */
 		hent = gethostbyname(LocalHostName);
-		if (hent != NULL)
-		{
+		if ( hent )
 			sprintf(LocalHostName, "%s", hent->h_name);
-			if ( (p = index(LocalHostName, '.')) )
-			{	
-				*p++ = '\0';
-				LocalDomain = p;
-			}
+			
+		if ( (p = index(LocalHostName, '.')) )
+		{
+			*p++ = '\0';
+			LocalDomain = p;
 		}
 	}
 
@@ -757,59 +847,6 @@ int main(argc, argv)
 	(void) alarm(TIMERINTVL);
 	(void) unlink(LogName);
 
-#ifdef SYSLOG_UNIXAF
-	sunx.sun_family = AF_UNIX;
-	(void) strncpy(sunx.sun_path, LogName, sizeof(sunx.sun_path));
-	funix = socket(AF_UNIX, SOCK_STREAM, 0);
-	if (funix < 0 || bind(funix, (struct sockaddr *) &sunx,
-	    sizeof(sunx.sun_family)+strlen(sunx.sun_path)) < 0 ||
-	    chmod(LogName, 0666) < 0 || listen(funix, 5) < 0) {
-		(void) sprintf(line, "cannot create %s", LogName);
-		logerror(line);
-		dprintf("cannot create %s (%d).\n", LogName, errno);
-#ifndef SYSV
-		die(0);
-#endif
-	}
-#endif
-
-#ifdef SYSLOG_INET
-	finet = socket(AF_INET, SOCK_DGRAM, 0);
-	if (finet >= 0) {
-	        auto int on = 1;
-		struct servent *sp;
-
-		sp = getservbyname("syslog", "udp");
-		if (sp == NULL) {
-			errno = 0;
-			logerror("network logging disabled (syslog/udp service unknown).");
-			logerror("see syslogd(8) for details of whether and how to enable it.");
-		}
-		else {
-			sin.sin_family = AF_INET;
-			sin.sin_port = LogPort = sp->s_port;
-			sin.sin_addr.s_addr = 0;
-			if ( setsockopt(finet, SOL_SOCKET, SO_REUSEADDR, \
-					(char *) &on, sizeof(on)) < 0 ) {
-				logerror("setsockopt, suspending inet");
-			}
-			else {
-				if (bind(finet, (struct sockaddr *) &sin, \
-					 sizeof(sin)) < 0) {
-					logerror("bind, suspending inet");
-				} else {
-					inetm = finet;
-					InetInuse = 1;
-					dprintf("listening on syslog UDP port.\n");
-				}
-			}
-		}
-	}
-	else
-		logerror("syslog: Unknown protocol, suspending inet service.");
-#endif
-
-
 	/* Create a partial message table for all file descriptors. */
 	num_fds = getdtablesize();
 	dprintf("Allocated parts table for %d file descriptors.\n", num_fds);
@@ -830,10 +867,6 @@ int main(argc, argv)
 		debugging_on = 0;
 	}
 
-	if (quitpid) {
-		kill(quitpid, SIGINT);
-	}
-
 	/* Main loop begins here. */
 	FD_ZERO(&unixm);
 	FD_ZERO(&readfds);
@@ -855,8 +888,10 @@ int main(argc, argv)
 		 * Add the Internet Domain Socket to the list of read
 		 * descriptors.
 		 */
-		if ( InetInuse && AcceptRemote )
+		if ( InetInuse && AcceptRemote ) {
 			FD_SET(inetm, &readfds);
+			dprintf("Listening on syslog UDP port.\n");
+		}
 #endif
 
 		if ( debugging_on )
@@ -906,7 +941,7 @@ int main(argc, argv)
 					dprintf("%d ", nfds);
 			dprintf("\n");
 		}
-		for (fd= 0; fd <= FD_SETSIZE; ++fd)
+		for (fd= 0; fd < FD_SETSIZE; ++fd)
 		  if ( FD_ISSET(fd, &readfds) && FD_ISSET(fd, &unixm) ) {
 			dprintf("Message from UNIX socket #%d.\n", fd);
 			memset(line, '\0', sizeof(line));
@@ -990,14 +1025,67 @@ int usage()
 	exit(1);
 }
 
+#ifdef SYSLOG_UNIXAF
+static int create_unix_socket()
+{
+	struct sockaddr_un sunx;
+	int fd;
+	char line[MAXLINE +1];
+
+	sunx.sun_family = AF_UNIX;
+	(void) strncpy(sunx.sun_path, LogName, sizeof(sunx.sun_path));
+	fd = socket(AF_UNIX, SOCK_STREAM, 0);
+	if (fd < 0 || bind(fd, (struct sockaddr *) &sunx,
+			   sizeof(sunx.sun_family)+strlen(sunx.sun_path)) < 0 ||
+	    chmod(LogName, 0666) < 0 || listen(fd, 5) < 0) {
+		(void) sprintf(line, "cannot create %s", LogName);
+		logerror(line);
+		dprintf("cannot create %s (%d).\n", LogName, errno);
+		close(fd);
+#ifndef SYSV
+		die(0);
+#endif
+	}
+	return fd;
+}
+#endif
+
+#ifdef SYSLOG_INET
+static int create_inet_socket()
+{
+	int fd, on = 1;
+	struct sockaddr_in sin;
+
+	fd = socket(AF_INET, SOCK_DGRAM, 0);
+	if (fd < 0) {
+		logerror("syslog: Unknown protocol, suspending inet service.");
+		return fd;
+	}
+
+	sin.sin_family = AF_INET;
+	sin.sin_port = LogPort;
+	sin.sin_addr.s_addr = 0;
+	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, \
+		       (char *) &on, sizeof(on)) < 0 ) {
+		logerror("setsockopt, suspending inet");
+		close(fd);
+		return -1;
+	}
+	if (bind(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
+		logerror("bind, suspending inet");
+		close(fd);
+		return -1;
+	}
+	return fd;
+}
+#endif
 
 char **
 crunch_list(list)
 	char *list;
 {
-	int count;
-	int i;
-	char *p;
+	int count, i;
+	char *p, *q;
 	char **result = NULL;
 
 	p = list;
@@ -1028,15 +1116,15 @@ crunch_list(list)
 	 * so we don't have to care about this.
 	 */
 	count = 0;
-	while ((i=(int)index(p, LIST_DELIMITER))) {
-		if ((result[count] = \
-		     (char *)malloc(sizeof(char) * i - (int)p +1)) == NULL) {
+	while ((q=index(p, LIST_DELIMITER))) {
+		result[count] = (char *) malloc((q - p + 1) * sizeof(char));
+		if (result[count] == NULL) {
 			printf ("Sorry, can't get enough memory, exiting.\n");
 			exit(0);
 		}
-		strncpy(result[count],p, i - (int)p);
-		result[count][i - (int)p] = '\0';
-		p = (char *)i;p++;
+		strncpy(result[count], p, q - p);
+		result[count][q - p] = '\0';
+		p = q; p++;
 		count++;
 	}
 	if ((result[count] = \
@@ -1239,6 +1327,24 @@ void printsys(msg)
 	return;
 }
 
+/*
+ * Decode a priority into textual information like auth.emerg.
+ */
+char *textpri(pri)
+	int pri;
+{
+	static char res[20];
+	CODE *c_pri, *c_fac;
+
+	for (c_fac = facilitynames; c_fac->c_name && !(c_fac->c_val == LOG_FAC(pri)<<3); c_fac++);
+	for (c_pri = prioritynames; c_pri->c_name && !(c_pri->c_val == LOG_PRI(pri)); c_pri++);
+
+	/*	sprintf (res, "%d.%d", LOG_FAC(pri), LOG_PRI(pri));*/
+	sprintf (res, "%s.%s<%d>", c_fac->c_name, c_pri->c_name, pri);
+
+	return res;
+}
+
 time_t	now;
 
 /*
@@ -1257,7 +1363,7 @@ void logmsg(pri, msg, from, flags)
 	int msglen;
 	char *timestamp;
 
-	dprintf("logmsg: pri %o, flags %x, from %s, msg %s\n", pri, flags, from, msg);
+	dprintf("logmsg: %s, flags %x, from %s, msg %s\n", textpri(pri), flags, from, msg);
 
 #ifndef SYSV
 	omask = sigblock(sigmask(SIGHUP)|sigmask(SIGALRM));
@@ -1346,13 +1452,13 @@ void logmsg(pri, msg, from, flags)
 			/* new line, save it */
 			if (f->f_prevcount)
 				fprintlog(f, (char *)from, 0, (char *)NULL);
+			f->f_prevpri = pri;
 			f->f_repeatcount = 0;
 			(void) strncpy(f->f_lasttime, timestamp, 15);
 			(void) strncpy(f->f_prevhost, from,
 					sizeof(f->f_prevhost));
 			if (msglen < MAXSVLINE) {
 				f->f_prevlen = msglen;
-				f->f_prevpri = pri;
 				(void) strcpy(f->f_prevline, msg);
 				fprintlog(f, (char *)from, flags, (char *)NULL);
 			} else {
@@ -1366,6 +1472,9 @@ void logmsg(pri, msg, from, flags)
 	(void) sigsetmask(omask);
 #endif
 }
+#if FALSE
+} /* balance parentheses for emacs */
+#endif
 
 void fprintlog(f, from, flags, msg)
 	register struct filed *f;
@@ -1459,6 +1568,7 @@ void fprintlog(f, from, flags, msg)
 			else {
 			        dprintf("%s found, resuming.\n", f->f_un.f_forw.f_hname);
 				bcopy(hp->h_addr, (char *) &f->f_un.f_forw.f_addr.sin_addr, hp->h_length);
+				f->f_prevcount = 0;
 				f->f_type = F_FORW;
 				goto f_forw;
 			}
@@ -1524,6 +1634,11 @@ void fprintlog(f, from, flags, msg)
 			v->iov_len = 1;
 		}
 	again:
+		/* f->f_file == -1 is an indicator that the we couldn't
+		   open the file at startup. */
+		if (f->f_file == -1)
+			break;
+
 		if (writev(f->f_file, iov, 6) < 0) {
 			int e = errno;
 
@@ -1573,6 +1688,9 @@ void fprintlog(f, from, flags, msg)
 		f->f_prevcount = 0;
 	return;		
 }
+#if FALSE
+}} /* balance parentheses for emacs */
+#endif
 
 jmp_buf ttybuf;
 
@@ -1812,7 +1930,7 @@ void logerror(type)
 {
 	char buf[100];
 
-	dprintf("Called loggerr, msg: %s\n", type);
+	dprintf("Called logerr, msg: %s\n", type);
 
 	if (errno == 0)
 		(void) sprintf(buf, "syslogd: %s", type);
@@ -1850,7 +1968,7 @@ void die(sig)
 
 	/* Close the sockets. */
         close(funix);
-	close(inetm);
+	if (InetInuse) close(inetm);
 
 	/* Clean-up files. */
 	(void) unlink(LogName);
@@ -1868,52 +1986,59 @@ void init()
 	register FILE *cf;
 	register struct filed *f, **nextp = (struct filed **) 0;
 	register char *p;
+	register unsigned int Forwarding = 0;
+#ifdef CONT_LINE
+	char cbuf[BUFSIZ];
+	char *cline;
+#else
 	char cline[BUFSIZ];
+#endif
+	struct servent *sp;
 
-	dprintf("Called init.\n");
+	sp = getservbyname("syslog", "udp");
+	if (sp == NULL) {
+		errno = 0;
+		logerror("network logging disabled (syslog/udp service unknown).");
+		logerror("see syslogd(8) for details of whether and how to enable it.");
+		return;
+	}
+	LogPort = sp->s_port;
 
 	/*
-	 *  Close all open log files.
+	 *  Close all open log files and free log descriptor array.
 	 */
-/*
+	dprintf("Called init.\n");
 	Initialized = 0;
 	if ( nlogs > -1 )
 	{
 		dprintf("Initializing log structures.\n");
+
+		for (lognum = 0; lognum <= nlogs; lognum++ ) {
+			f = &Files[lognum];
+
+			/* flush any pending output */
+			if (f->f_prevcount)
+				fprintlog(f, LocalHostName, 0, (char *)NULL);
+
+			switch (f->f_type) {
+				case F_FILE:
+				case F_PIPE:
+				case F_TTY:
+				case F_CONSOLE:
+					(void) close(f->f_file);
+				break;
+			}
+		}
+
+		/*
+		 * This is needed especially when HUPing syslogd as the
+		 * structure would grow infinitively.  -Joey
+		 */
 		nlogs = -1;
 		free((void *) Files);
 		Files = (struct filed *) 0;
 	}
-*/
 	
-#ifdef SYSV
-	for (lognum = 0; lognum <= nlogs; lognum++ ) {
-		f = &Files[lognum];
-#else
-	for (f = Files; f != NULL; f = next) {
-#endif
-		/* flush any pending output */
-		if (f->f_prevcount)
-			fprintlog(f, LocalHostName, 0, (char *)NULL);
-
-		switch (f->f_type) {
-		  case F_FILE:
-		  case F_PIPE:
-		  case F_TTY:
-		  case F_CONSOLE:
-			(void) close(f->f_file);
-			break;
-		}
-#ifdef SYSV
-		f->f_type = F_UNUSED;	/* clear entry - ASP */
-	}
-#else
-		next = f->f_next;
-		free((char *) f);
-	}
-	Files = NULL;
-	nextp = &OBFiles;
-#endif
 
 	/* open the configuration file */
 	if ((cf = fopen(ConfFile, "r")) == NULL) {
@@ -1938,7 +2063,12 @@ void init()
 #else
 	f = NULL;
 #endif
+#if CONT_LINE
+	cline = cbuf;
+	while (fgets(cline, sizeof(cbuf) - (cline - cbuf), cf) != NULL) {
+#else
 	while (fgets(cline, sizeof(cline), cf) != NULL) {
+#endif
 		/*
 		 * check for end-of-section, comments, strip off trailing
 		 * spaces and newline character.
@@ -1946,7 +2076,23 @@ void init()
 		for (p = cline; isspace(*p); ++p);
 		if (*p == '\0' || *p == '#')
 			continue;
+#if CONT_LINE
+		strcpy(cline, p);
+#endif
 		for (p = index(cline, '\0'); isspace(*--p););
+#if CONT_LINE
+		if (*p == '\\') {
+			if ((p - cbuf) > BUFSIZ - 30) {
+				/* Oops the buffer is full - what now? */
+				cline = cbuf;
+			} else {
+				*p = 0;
+				cline = p;
+				continue;
+			}
+		}  else
+			cline = cbuf;
+#endif
 		*++p = '\0';
 #ifndef SYSV
 		f = (struct filed *)calloc(1, sizeof(*f));
@@ -1955,12 +2101,43 @@ void init()
 #endif
 		allocate_log();
 		f = &Files[lognum++];
+#if CONT_LINE
+		cfline(cbuf, f);
+#else
 		cfline(cline, f);
+#endif
+		if (f->f_type == F_FORW || f->f_type == F_FORW_SUSP || f->f_type == F_FORW_UNKN) {
+			Forwarding++;
+		}
 	}
 
 	/* close the configuration file */
 	(void) fclose(cf);
 
+#ifdef SYSLOG_UNIXAF
+	if (funix < 0)
+		funix = create_unix_socket();
+#endif
+
+#ifdef SYSLOG_INET
+	if (Forwarding || AcceptRemote) {
+		if (finet < 0) {
+			finet = create_inet_socket();
+			if (finet >= 0) {
+				InetInuse = 1;
+				dprintf("Opened syslog UDP port.\n");
+			}
+		}
+	}
+	else {
+		if (finet >= 0)
+			close(finet);
+		finet = -1;
+		InetInuse = 0;
+	}
+	inetm = finet;
+#endif
+
 	Initialized = 1;
 
 	if ( Debug ) {
@@ -1985,6 +2162,8 @@ void init()
 				case F_TTY:
 				case F_CONSOLE:
 					printf("%s", f->f_un.f_fname);
+					if (f->f_file == -1)
+						printf(" (unused)");
 					break;
 
 				case F_FORW:
@@ -2004,15 +2183,29 @@ void init()
 	}
 
 	if ( AcceptRemote )
+#ifdef DEBRELEASE
 		logmsg(LOG_SYSLOG|LOG_INFO, "syslogd " VERSION "-" PATCHLEVEL "#" DEBRELEASE \
 		       ": restart (remote reception)." , LocalHostName, \
 		       	ADDDATE);
+#else
+		logmsg(LOG_SYSLOG|LOG_INFO, "syslogd " VERSION "-" PATCHLEVEL \
+		       ": restart (remote reception)." , LocalHostName, \
+		       	ADDDATE);
+#endif
 	else
+#ifdef DEBRELEASE
 		logmsg(LOG_SYSLOG|LOG_INFO, "syslogd " VERSION "-" PATCHLEVEL "#" DEBRELEASE \
 		       ": restart." , LocalHostName, ADDDATE);
+#else
+		logmsg(LOG_SYSLOG|LOG_INFO, "syslogd " VERSION "-" PATCHLEVEL \
+		       ": restart." , LocalHostName, ADDDATE);
+#endif
 	(void) signal(SIGHUP, sighup_handler);
 	dprintf("syslogd: restarted.\n");
 }
+#if FALSE
+	}}} /* balance parentheses for emacs */
+#endif
 
 /*
  * Crack a configuration file line
@@ -2069,6 +2262,9 @@ void cfline(line, f)
 				*bp=*(bp+1);
 			*bp='\0';
 		}
+		else {
+			ignorepri = 0;
+		}
 		if ( *buf == '=' )
 		{
 			singlpri = 1;
@@ -2093,7 +2289,7 @@ void cfline(line, f)
 				*bp++ = *p++;
 			*bp = '\0';
 			if (*buf == '*') {
-				for (i = 0; i < LOG_NFACILITIES; i++) {
+				for (i = 0; i <= LOG_NFACILITIES; i++) {
 					if ( pri == INTERNAL_NOPRI ) {
 						if ( ignorepri )
 							f->f_pmask[i] = TABLE_ALLPRI;
@@ -2184,16 +2380,16 @@ void cfline(line, f)
 	{
 	case '@':
 #ifdef SYSLOG_INET
-		if (!InetInuse)
-			break;
 		(void) strcpy(f->f_un.f_forw.f_hname, ++p);
 		dprintf("forwarding host: %s\n", p);	/*ASP*/
 		if ( (hp = gethostbyname(p)) == NULL ) {
 			f->f_type = F_FORW_UNKN;
 			f->f_prevcount = INET_RETRY_MAX;
+			f->f_time = time ( (time_t *)0 );
 		} else {
 			f->f_type = F_FORW;
 		}
+
 		bzero((char *) &f->f_un.f_forw.f_addr,
 			 sizeof(f->f_un.f_forw.f_addr));
 		f->f_un.f_forw.f_addr.sin_family = AF_INET;
@@ -2225,7 +2421,7 @@ void cfline(line, f)
 		}
 		        
 	  	if ( f->f_file < 0 ){
-			f->f_file = F_UNUSED;
+			f->f_file = -1;
 			dprintf("Error opening log file: %s\n", p);
 			logerror(p);
 			break;
@@ -2274,7 +2470,7 @@ int decode(name, codetab)
 {
 	register struct code *c;
 	register char *p;
-	char buf[40];
+	char buf[80];
 
 	dprintf ("symbolic name: %s", name);
 	if (isdigit(*name))
@@ -2282,7 +2478,7 @@ int decode(name, codetab)
 		dprintf ("\n");
 		return (atoi(name));
 	}
-	(void) strcpy(buf, name);
+	(void) strncpy(buf, name, 79);
 	for (p = buf; *p; p++)
 		if (isupper(*p))
 			*p = tolower(*p);
@@ -2371,3 +2567,12 @@ void sighup_handler()
 	signal(SIGHUP, sighup_handler);
 	return;
 }
+
+/*
+ * Local variables:
+ *  c-indent-level: 8
+ *  c-basic-offset: 8
+ *  tab-width: 8
+ * End:
+ */
+
